import abc


class AbstractProvider(abc.ABC):

    @abc.abstractmethod
    async def fetch_messages(self):
        '''A coroutine that returns a sequence of messages to be processed.
        If no messages are available, this coroutine should return an empty list.
        '''

    @abc.abstractmethod
    async def confirm_message(self, message):
        '''A coroutine to confirm the message processing.
        After the message confirmation we should not receive the same message again.
        This usually means we need to delete the message in the provider.
        '''

    def stop(self):
        '''Stops the provider.
        If needed, the provider should perform clean-up actions.
        This method is called whenever we need to shutdown the provider.
        '''
        pass



import asyncio
import logging

import aiobotocore
from cached_property import cached_property

logger = logging.getLogger(__name__)


class BaseSQSClient:

    def __init__(self, endpoint_url=None, use_ssl=True, loop=None):
        self.endpoint_url = endpoint_url
        self.use_ssl = use_ssl
        self._loop = loop or asyncio.get_event_loop()
        self._cached_queue_urls = {}

    @cached_property
    def client(self):
        session = aiobotocore.get_session(loop=self._loop)
        return session.create_client('sqs', endpoint_url=self.endpoint_url, use_ssl=self.use_ssl)

    async def get_queue_url(self, queue):
        if queue and (queue.startswith('http://') or queue.startswith('https://')):
            name = queue.split('/')[-1]
            self._cached_queue_urls[name] = queue
            queue = name

        if queue not in self._cached_queue_urls:
            response = await self.client.get_queue_url(QueueName=queue)
            self._cached_queue_urls[queue] = response['QueueUrl']

        return self._cached_queue_urls[queue]

    def stop(self):
        logger.info('closing client={}'.format(self.client))
        self.client.close()


class BaseSNSClient:

    def __init__(self, endpoint_url=None, use_ssl=True, loop=None):
        self.endpoint_url = endpoint_url
        self.use_ssl = use_ssl
        self._loop = loop or asyncio.get_event_loop()

    @cached_property
    def client(self):
        session = aiobotocore.get_session(loop=self._loop)
        return session.create_client('sns', endpoint_url=self.endpoint_url, use_ssl=self.use_ssl)

    async def get_topic_arn(self, topic):
        if topic.startswith('arn:sns:'):
            return topic
        return 'arn:sns:*:{}'.format(topic)

    def stop(self):
        logger.info('closing client={}'.format(self.client))
        self.client.close()



from functools import wraps
import importlib
import os
import sys


def add_current_dir_to_syspath(f):

    @wraps(f)
    def wrapper(*args, **kwargs):
        current = os.getcwd()
        changed = False
        if current not in sys.path:
            sys.path.append(current)
            changed = True

        try:
            return f(*args, **kwargs)
        finally:
            # restore sys.path
            if changed is True:
                sys.path.remove(current)

    return wrapper


@add_current_dir_to_syspath
def import_callable(full_name):
    package, *name = full_name.rsplit('.', 1)
    try:
        module = importlib.import_module(package)
    except ValueError as exc:
        raise ImportError('Error trying to import {!r}'.format(full_name)) from exc

    if name:
        handler = getattr(module, name[0])
    else:
        handler = module

    if not callable(handler):
        raise ImportError('{!r} should be callable'.format(full_name))

    return handler



import logging

import botocore.exceptions

from loafer.exceptions import ProviderError
from loafer.providers import AbstractProvider
from .bases import BaseSQSClient

logger = logging.getLogger(__name__)


class SQSProvider(AbstractProvider, BaseSQSClient):

    def __init__(self, queue_name, options=None, **kwargs):
        self.queue_name = queue_name
        self._options = options or {}
        super().__init__(**kwargs)

    def __str__(self):
        return '<{}: {}>'.format(type(self).__name__, self.queue_name)

    async def confirm_message(self, message):
        logger.info('confirm message (ack/deletion)')

        receipt = message['ReceiptHandle']
        logger.debug('receipt={!r}'.format(receipt))

        queue_url = await self.get_queue_url(self.queue_name)
        return await self.client.delete_message(QueueUrl=queue_url, ReceiptHandle=receipt)

    async def fetch_messages(self):
        logger.debug('fetching messages on {}'.format(self.queue_name))
        try:
            queue_url = await self.get_queue_url(self.queue_name)
            response = await self.client.receive_message(QueueUrl=queue_url, **self._options)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as exc:
            raise ProviderError('error fetching messages') from exc

        return response.get('Messages', [])

    def stop(self):
        logger.info('stopping {}'.format(self))
        self.client.close()



import asyncio
import logging

from .message_translators import AbstractMessageTranslator
from .providers import AbstractProvider

logger = logging.getLogger(__name__)


class Route:

    def __init__(self, provider, handler, name='default',
                 message_translator=None, error_handler=None):
        self.name = name

        assert isinstance(provider, AbstractProvider), 'invalid provider instance'
        self.provider = provider

        self.message_translator = message_translator
        if message_translator:
            assert isinstance(message_translator, AbstractMessageTranslator), \
                'invalid message translator instance'

        self._error_handler = error_handler
        if error_handler:
            assert callable(error_handler), 'error_handler must be a callable object'

        if callable(handler):
            self.handler = handler
            self._handler_instance = None
        else:
            self.handler = getattr(handler, 'handle', None)
            self._handler_instance = handler

        assert self.handler, 'handler must be a callable object or implement `handle` method'

    def __str__(self):
        return '<{}(name={} provider={!r} handler={!r})>'.format(
            type(self).__name__, self.name, self.provider, self.handler)

    def apply_message_translator(self, message):
        processed_message = {'content': message,
                             'metadata': {}}
        if not self.message_translator:
            return processed_message

        translated = self.message_translator.translate(processed_message['content'])
        processed_message['metadata'].update(translated.get('metadata', {}))
        processed_message['content'] = translated['content']
        if not processed_message['content']:
            raise ValueError('{} failed to translate message={}'.format(self.message_translator, message))

        return processed_message

    async def deliver(self, raw_message, loop=None):
        message = self.apply_message_translator(raw_message)
        logger.info(
            'delivering message content to handler={!r}, message={!r}'.format(self.handler, message)
        )

        if asyncio.iscoroutinefunction(self.handler):
            logger.debug('handler is coroutine! {!r}'.format(self.handler))
            return await self.handler(message['content'], message['metadata'])
        else:
            logger.debug('handler will run in a separate thread: {!r}'.format(self.handler))
            loop = loop or asyncio.get_event_loop()
            return await loop.run_in_executor(None, self.handler, message['content'], message['metadata'])

    async def error_handler(self, exc_info, message, loop=None):
        logger.info('error handler process originated by message={}'.format(message))
        if self._error_handler is not None:
            if asyncio.iscoroutinefunction(self._error_handler):
                return await self._error_handler(exc_info, message)
            else:
                loop = loop or asyncio.get_event_loop()
                return await loop.run_in_executor(None, self._error_handler, exc_info, message)

        return False

    def stop(self):
        logger.info('stopping route {}'.format(self))
        self.provider.stop()
        # only for class-based handlers
        if hasattr(self._handler_instance, 'stop'):
            self._handler_instance.stop()



import abc
import logging

logger = logging.getLogger(__name__)


class AbstractMessageTranslator(abc.ABC):

    @abc.abstractmethod
    def translate(self, message):
        '''Translates a given message to an appropriate format to message processing.
        This method should return a `dict` instance with two keys: `content`
        and `metadata`.
        The `content` should contain the translated message and, `metadata` a
        dictionary with translation metadata or an empty `dict`.
        '''


class StringMessageTranslator(AbstractMessageTranslator):

    def translate(self, message):
        logger.debug('{!r} will translate {!r}'.format(type(self).__name__, message))
        return {'content': str(message), 'metadata': {}}



import json
import logging

from .bases import BaseSQSClient, BaseSNSClient

logger = logging.getLogger(__name__)


class SQSHandler(BaseSQSClient):
    queue_name = None

    def __init__(self, queue_name=None, **kwargs):
        self.queue_name = queue_name or self.queue_name
        super().__init__(**kwargs)

    def __str__(self):
        return '<{}: {}>'.format(type(self).__name__, self.queue_name)

    async def publish(self, message, encoder=json.dumps):
        if not self.queue_name:
            raise ValueError('{}: missing queue_name attribute'.format(type(self).__name__))

        if encoder:
            message = encoder(message)

        logger.debug('publishing, queue={}, message={}'.format(self.queue_name, message))

        queue_url = await self.get_queue_url(self.queue_name)
        return await self.client.send_message(QueueUrl=queue_url, MessageBody=message)

    async def handle(self, message, *args):
        return await self.publish(message)


class SNSHandler(BaseSNSClient):
    topic = None

    def __init__(self, topic=None, **kwargs):
        self.topic = topic or self.topic
        super().__init__(**kwargs)

    def __str__(self):
        return '<{}: {}>'.format(type(self).__name__, self.topic)

    async def publish(self, message, encoder=json.dumps):
        if not self.topic:
            raise ValueError('{}: missing topic attribute'.format(type(self).__name__))

        if encoder:
            message = encoder(message)

        topic_arn = await self.get_topic_arn(self.topic)
        logger.debug('publishing, topic={}, message={}'.format(topic_arn, message))

        msg = json.dumps({'default': message})
        return await self.client.publish(TopicArn=topic_arn, MessageStructure='json', Message=msg)

    async def handle(self, message, *args):
        return await self.publish(message)



from ...routes import Route
from .providers import SQSProvider
from .message_translators import SQSMessageTranslator, SNSMessageTranslator


class SQSRoute(Route):
    def __init__(self, provider_queue, provider_options=None, *args, **kwargs):
        provider_options = provider_options or {}
        provider = SQSProvider(provider_queue, **provider_options)
        kwargs['provider'] = provider
        if 'message_translator' not in kwargs:
            kwargs['message_translator'] = SQSMessageTranslator()
        if 'name' not in kwargs:
            kwargs['name'] = provider_queue

        super().__init__(*args, **kwargs)


class SNSQueueRoute(Route):
    def __init__(self, provider_queue, provider_options=None, *args, **kwargs):
        provider_options = provider_options or {}
        provider = SQSProvider(provider_queue, **provider_options)
        kwargs['provider'] = provider
        if 'message_translator' not in kwargs:
            kwargs['message_translator'] = SNSMessageTranslator()
        if 'name' not in kwargs:
            kwargs['name'] = provider_queue

        super().__init__(*args, **kwargs)



import codecs
import os.path
import re
from setuptools import setup, find_packages, Command


# metadata

here = os.path.abspath(os.path.dirname(__file__))
version = "0.0.0"
changes = os.path.join(here, "CHANGES.rst")
pattern = r'^(?P<version>[0-9]+.[0-9]+(.[0-9]+)?)'
with codecs.open(changes, encoding='utf-8') as changes:
    for line in changes:
        match = re.match(pattern, line)
        if match:
            version = match.group("version")
            break


class VersionCommand(Command):
    description = 'Show library version'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print(version)


with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = '\n{}'.format(f.read())

with codecs.open(os.path.join(here, 'CHANGES.rst'), encoding='utf-8') as f:
    changes = f.read()
    long_description += '\n\nChangelog:\n----------\n\n{}'.format(changes)


# Requirements

# Unduplicated tests_requirements and requirements/test.txt
tests_requirements = [
    'pytest',
    'pytest-asyncio',
    'pytest-cov',
    'coveralls',
    'asynctest',
    'flake8',
    'tox'
]

# aiohttp version are directly tied with aiobotocore, please check before update it.
install_requirements = [
    'aiohttp>=2.0.4',
    'boto3>=1.3.0',
    'aiobotocore>=0.3.0',
    'cached-property>=1.3.0',
]


# setup

setup(
    name='loafer',
    version=version,
    description='Asynchronous message dispatcher for concurrent tasks processing',
    long_description=long_description,
    url='https://github.com/georgeyk/loafer/',
    download_url='https://github.com/georgeyk/loafer/releases',
    license='MIT',
    author='George Y. Kussumoto',
    author_email='contato@georgeyk.com.br',
    packages=find_packages(exclude=['docs', 'tests', 'tests.*', 'requirements']),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: System :: Distributed Computing',
    ],
    keywords='asynchronous asyncio message dispatcher tasks microservices',
    setup_requires=['pytest-runner'],
    install_requires=install_requirements,
    tests_require=tests_requirements,
    cmdclass={'version': VersionCommand},
)



# TODO: it should be async
# https://github.com/getsentry/raven-aiohttp


def sentry_handler(client, delete_message=False):

    def send_to_sentry(exc_info, message):
        client.captureException(
            exc_info,
            extra={'message': message},
        )
        return delete_message

    return send_to_sentry
